import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from tqdm import trange

from . import Model as SPM



def OriginalCauchyMethod(H0, W0, nmbMaxPoles=40, dWorigin=1e-3, phys=True,
                         useLogPrec=False, logPrec=15, plotSingularValues=False, figID=11111):
    # Adjust some variables
    W1, H1 = None, None
    if not phys:
        W1, H1 = W0, H0
    else:
        ids1 = np.arange(0, W0.shape[0], 2)
        ids2 = np.arange(1, W0.shape[0], 2)
        W1 = np.hstack([-np.flip(W0[ids1]), W0[ids2]])
        H1 = np.hstack([np.flip(H0[ids1]).conj(), H0[ids2]])
    H = H1[..., np.newaxis]
    nmbMaxPoles = min(nmbMaxPoles, W0.shape[0])
    e0 = (np.abs(H0) ** 2).mean()

    # Set dQ and dP
    Mp = nmbMaxPoles
    Mz = Mp - 1

    # First SVD to find the singular values
    Wd = np.vander(W1, N=Mp + 1, increasing=True)
    B = H * Wd
    A = np.vander(W1, N=Mz + 1, increasing=True)
    C = np.hstack([A, -B])
    U, S, Vh = np.linalg.svd(C)

    # Look at the log ratio of the singular values S and the max singular value, and the growth of S
    yS = np.log10(S / np.max(S))
    yDiffS = np.abs(yS[1:] - yS[:-1])

    # Define the max rank using either the LogPrec or the growth stop point
    rsMax = nmbMaxPoles
    if useLogPrec:
        rsMax = np.argwhere(yS > -logPrec)[-1, 0]
        rsMax = max(rsMax, 4)  # diffZP+3 = Mz:1 + Mp:1+diffZP + 2 - K:1
    else:
        rsMax = np.argwhere(yDiffS == 0)
        if len(rsMax) > 0:
            rsMax = rsMax[0, 0] + 1
        else:
            rsMax = np.argwhere(yS > -logPrec)[-1, 0] + 1
        rsMax = max(rsMax, 4)  # diffZP+3 = Mz:1 + Mp:1+diffZP + 2 - K:1

    # Plot the singular values if you want...
    if plotSingularValues:
        fig = plt.figure(figID, figsize=[5, 4])
        plt.clf()
        fig.subplots_adjust(left=.17, bottom=.15, right=.85)
        ax = fig.subplots(1)
        ax.plot(np.arange(S.shape[0]) + 1, yS, color="black", linewidth=2.4)
        ax.scatter(x=np.arange(S.shape[0]) + 1, y=yS, s=13, color="blue", marker="x")
        ax.set_xlabel("Index $i$ of singular value", fontsize=12)
        ax.set_ylabel(r"log[$\sigma_i / \sigma_{max}$]", fontsize=12)
        ax.axvline(rsMax, color="black", linestyle="--")
        ax2 = ax.twinx()
        ax2.plot(np.arange(S.shape[0] - 1) + 1, yDiffS, linestyle="--", color="red")
        ax2.set_ylabel("Derivative approximation", color="red", fontsize=12)
        fig.show()

    g0, p, z = None, None, None
    if rsMax >= 4:
        Mp = rsMax // 2
        Mz = Mp - 1
        if not (Mp + Mz + 1 == rsMax):
            Mp = (rsMax + 1) // 2
            Mz = Mp - 1
        A = np.vander(W1, N=Mz + 1, increasing=True)
        Wd = np.vander(W1, N=Mp + 1, increasing=True)
        B = H * Wd
        C = np.hstack([A, -B])

        U, S, Vh = np.linalg.svd(C, full_matrices=False)
        S[-1] = 0
        C = np.dot(U * S, Vh)
        A = C[:, :(Mz + 1)]
        B = -C[:, (Mz + 1):]

        Q, R1 = np.linalg.qr(A, mode="complete")
        R11 = R1[:(Mz + 1), :(Mz + 1)]

        R2 = -np.matmul(Q.T, B)
        R21 = R2[:(Mz + 1), :]
        R22 = R2[(Mz + 1):, :]

        U, S, V = np.linalg.svd(R22, full_matrices=True)
        b = V.conj().T[:, -1]
        a = -np.matmul(np.linalg.inv(R11), np.matmul(R21, b))

        g0 = np.abs(a[-1] / b[-1]) if phys else a[-1] / b[-1]
        pa = np.polynomial.Polynomial(a)
        pb = np.polynomial.Polynomial(b)
        z = pa.roots()
        p = pb.roots()

        if (z.shape[0] > 0) and (p.shape[0] > 0):
            zIm = z[np.abs(np.real(z)) < dWorigin]
            zPairs = z[np.abs(np.real(z)) >= dWorigin]
            if zPairs.shape[0] > 0:
                zPairs = zPairs[np.real(zPairs) > -dWorigin]

            pIm = p[np.abs(np.real(p)) < dWorigin]
            pPairs = p[np.abs(np.real(p)) >= dWorigin]
            if pPairs.shape[0] > 0:
                pPairs = pPairs[np.real(pPairs) > -dWorigin]

            zeros = np.hstack([-np.flip(zPairs).conj(), 1j * np.imag(zIm), zPairs]) if phys else z
            poles = np.hstack([-np.flip(pPairs).conj(), 1j * np.imag(pIm), pPairs]) if phys else p
            Hwpred = SPM.G_SZF(W0, g0, poles, zeros)
            p, z = poles, zeros

    return g0, p, z


def CauchyMethod(H0, W0, nmbMaxPoles=40, dWorigin=1e-3, phys=True,
                 useLogPrec=False, logPrec=15, diffZP=0, plotSingularValues=False, figID=11111):
    # Adjust some variables
    W1, H1 = None, None
    if not phys:
        W1, H1 = W0, H0
    else:
        ids1 = np.arange(0, W0.shape[0], 2)
        ids2 = np.arange(1, W0.shape[0], 2)
        W1 = np.hstack([-np.flip(W0[ids1]), W0[ids2]])
        H1 = np.hstack([np.flip(H0[ids1]).conj(), H0[ids2]])
    H = H1[..., np.newaxis]
    nmbMaxPoles = min(nmbMaxPoles, W0.shape[0])
    e0 = (np.abs(H0) ** 2).mean()

    # Set dQ and dP
    Mp = nmbMaxPoles
    Mz = max(1, nmbMaxPoles - diffZP)

    # First SVD to find the singular values
    Wd = np.vander(W1, N=Mp + 1, increasing=True)
    B = H * Wd
    A = np.vander(W1, N=Mz + 1, increasing=True)
    C = np.hstack([A, -B])
    U, S, Vh = np.linalg.svd(C)

    # Look at the log ratio of the singular values S and the max singular value, and the growth of S
    yS = np.log10(S / np.max(S))
    yDiffS = np.abs(yS[1:] - yS[:-1])

    # Define the max rank using either the LogPrec or the growth stop point
    rsMax = nmbMaxPoles
    if useLogPrec:
        rsMax = np.argwhere(yS > -logPrec)[-1, 0]
        rsMax = max(rsMax, diffZP + 3)  # diffZP+3 = Mz:1 + Mp:1+diffZP + 2 - K:1
    else:
        rsMax = np.argwhere(yDiffS == 0)
        if len(rsMax) > 0:
            rsMax = rsMax[0, 0] + 1
        else:
            rsMax = np.argwhere(yS > -logPrec)[-1, 0] + 1
        rsMax = max(rsMax, diffZP + 3)  # diffZP+3 = Mz:1 + Mp:1+diffZP + 2 - K:1

    # Plot the singular values if you want...
    if plotSingularValues:
        fig = plt.figure(figID, figsize=[5, 4])
        plt.clf()
        fig.subplots_adjust(left=.16, bottom=.15, right=.85)
        ax = fig.subplots(1)
        ax.plot(np.arange(S.shape[0]) + 1, yS, color="black")
        ax.scatter(x=np.arange(S.shape[0]) + 1, y=yS, s=7, color="blue", marker="x")
        ax.set_xlabel("Index $i$ of singular value")
        ax.set_ylabel(r"log[$\sigma_i / \sigma_{max}$]")
        ax.axvline(rsMax, color="black", linestyle="--")
        ax2 = ax.twinx()
        ax2.plot(np.arange(S.shape[0] - 1) + 1, yDiffS, linestyle="--", color="red")
        ax2.set_ylabel("Derivative approximation", color="red")
        fig.show()

    # Now start from that max rank and add poles and zeros until the error is low
    g0n = []
    zn = []
    pn = []
    Yn = []
    error = []
    rsArray = []
    # rs = diffZP+3 + 30
    if rsMax >= (diffZP + 3):
        Mp0 = (rsMax + diffZP - 1) // 2
        Mz0 = Mp - diffZP
        if not (Mp0 + Mz0 + 1 == rsMax):
            Mp0 = (rsMax + 1 + diffZP - 1) // 2
            Mz0 = Mp0 - diffZP
        A0 = np.vander(W1, N=Mz0 + 1, increasing=True)
        Wd = np.vander(W1, N=Mp0 + 1, increasing=True)
        B0 = H * Wd
        C0 = np.hstack([A0, -B0])

        dimMin = 3
        dimMax = Mz0 - 1
        for i in range(dimMin, dimMax):
            g0 = None
            zeros = None
            poles = None
            Hwpred = None
            e = np.inf

            Mz = i + 1
            Mp = Mz + diffZP
            rs = Mz + Mp + 3

            dimC = np.hstack([np.arange(Mz + 1), np.arange(Mz0 + 1, Mz0 + 1 + Mp + 1)])
            C = C0[:, dimC]

            U, S, Vh = np.linalg.svd(C, full_matrices=False)
            S[-1] = 0
            C = np.dot(U * S, Vh)
            A = C[:, :(Mz + 1)]
            B = -C[:, (Mz + 1):]

            Q, R1 = np.linalg.qr(A, mode="complete")
            R11 = R1[:(Mz + 1), :(Mz + 1)]

            R2 = -np.matmul(Q.T, B)
            R21 = R2[:(Mz + 1), :]
            R22 = R2[(Mz + 1):, :]

            U, S, V = np.linalg.svd(R22)
            b = V.conj().T[:, -1]
            a = -np.matmul(np.linalg.inv(R11), np.matmul(R21, b))

            g0 = np.abs(a[-1] / b[-1]) if phys else a[-1] / b[-1]
            pa = np.polynomial.Polynomial(a)
            pb = np.polynomial.Polynomial(b)
            z = pa.roots()
            p = pb.roots()

            if (z.shape[0] > 0) and (p.shape[0] > 0):
                zIm = z[np.abs(np.real(z)) < dWorigin]
                zPairs = z[np.abs(np.real(z)) >= dWorigin]
                if zPairs.shape[0] > 0:
                    zPairs = zPairs[np.real(zPairs) > -dWorigin]

                pIm = p[np.abs(np.real(p)) < dWorigin]
                pPairs = p[np.abs(np.real(p)) >= dWorigin]
                if pPairs.shape[0] > 0:
                    pPairs = pPairs[np.real(pPairs) > -dWorigin]

                zeros = np.hstack([-np.flip(zPairs).conj(), 1j * np.imag(zIm), zPairs]) if phys else z
                poles = np.hstack([-np.flip(pPairs).conj(), 1j * np.imag(pIm), pPairs]) if phys else p
                Hwpred = SPM.G_SZF(W0, g0, poles, zeros)
                p, z = poles, zeros
                e = np.abs((Hwpred - H0) ** 2).mean() / e0 * 100

            rsArray.append(rs)
            error.append(e)
            g0n.append(g0)
            zn.append(z)
            pn.append(p)
            Yn.append(Hwpred)

    idMin = np.argmin(error)
    g0 = g0n[idMin]
    zeros = zn[idMin]
    poles = pn[idMin]

    return g0, poles, zeros


def CauchyMethodOptZP(H0, W0, nmbMaxPoles=40, dWorigin=1e-3, phys=True,
                      stability=False, qStability=.1,
                      useLogPrec=False, logPrec=15, diffZPMax=5, plotSingularValues=False, figID=11111):
    # Adjust some variables
    W1, H1 = None, None
    if not phys:
        W1, H1 = W0, H0
    else:
        # ids1 = np.arange(0, W0.shape[0], 2)
        # ids2 = np.arange(1, W0.shape[0], 2)
        # W1 = np.hstack([-np.flip(W0[ids1]), W0[ids2]])
        # H1 = np.hstack([np.flip(H0[ids1]).conj(), H0[ids2]])
        W1 = np.hstack([-np.flip(W0), W0])
        H1 = np.hstack([np.flip(H0).conj(), H0])
    H = H1[..., np.newaxis]
    nmbMaxPoles = min(nmbMaxPoles, W0.shape[0])
    e0 = (np.abs(H0) ** 2).mean()

    # Set dQ and dP
    Mp = max(nmbMaxPoles, diffZPMax + 1)
    Mz = Mp

    # First SVD to find the singular values
    Wd = np.vander(W1, N=Mp + 1, increasing=True)
    B = H * Wd
    A = np.vander(W1, N=Mz + 1, increasing=True)
    C = np.hstack([A, -B])
    U, S, Vh = np.linalg.svd(C)

    # Look at the log ratio of the singular values S and the max singular value, and the growth of S
    yS = np.log10(S / np.max(S))
    yDiffS = np.abs(yS[1:] - yS[:-1])

    # Define the max rank using either the LogPrec or the growth stop point
    rsMax = nmbMaxPoles
    if useLogPrec:
        rsMax = np.argwhere(yS > -logPrec)[-1, 0]
        rsMax = max(rsMax, diffZPMax + 3)  # diffZP+3 = Mz:1 + Mp:1+diffZP + 2 - K:1
    else:
        rsMax = np.argwhere(yDiffS == 0)
        if len(rsMax) > 0:
            rsMax = rsMax[0, 0] + 1
        else:
            rsMax = np.argwhere(yS > -logPrec)[-1, 0] + 1
        rsMax = max(rsMax, diffZPMax + 3)  # diffZP+3 = Mz:1 + Mp:1+diffZP + 2 - K:1

    # Plot the singular values if you want...
    if plotSingularValues:
        fig = plt.figure(figID, figsize=[5, 4])
        plt.clf()
        fig.subplots_adjust(left=.16, bottom=.15, right=.85)
        ax = fig.subplots(1)
        ax.plot(np.arange(S.shape[0]) + 1, yS, color="black")
        ax.scatter(x=np.arange(S.shape[0]) + 1, y=yS, s=7, color="blue", marker="x")
        ax.set_xlabel("Index $i$ of singular value")
        ax.set_ylabel(r"log[$\sigma_i / \sigma_{max}$]")
        ax.axvline(rsMax, color="black", linestyle="--")
        ax2 = ax.twinx()
        ax2.plot(np.arange(S.shape[0] - 1) + 1, yDiffS, linestyle="--", color="red")
        ax2.set_ylabel("Derivative approximation", color="red")
        fig.show()

    # Now start from that max rank and add poles and zeros until the error is low
    g0n = []
    zn = []
    pn = []
    error = []
    Mzn = []
    Mpn = []

    rsMax += 6
    rsMax = min(rsMax, 2 * ((W1.shape[0] - 1) // 2))
    if not (rsMax % 2):
        rsMax += 1
    Mp0 = rsMax // 2
    Mz0 = Mp0
    A0 = np.vander(W1, N=Mz0 + 1, increasing=True)
    Wd = np.vander(W1, N=Mp0 + 1, increasing=True)
    B0 = H * Wd
    C0 = np.hstack([A0, -B0])

    for Mz in range(1, Mz0 + 1):
        for Mp in range(Mz, Mp0 + 1):
            Mzn.append(Mz)
            Mpn.append(Mp)

            g0 = None
            zeros = None
            poles = None
            Hwpred = None
            e = np.inf

            dimC = np.hstack([np.arange(Mz + 1), np.arange(Mz0 + 1, Mz0 + 1 + Mp + 1)])
            C = C0[:, dimC]

            # print(C.shape)
            U, S, Vh = np.linalg.svd(C, full_matrices=False)
            S[-1] = 0
            C = np.dot(U * S, Vh)
            A = C[:, :(Mz + 1)]
            B = -C[:, (Mz + 1):]
            # print(C.shape)

            Q, R1 = np.linalg.qr(A, mode="complete")
            R11 = R1[:(Mz + 1), :(Mz + 1)]
            #             print(Mz, Mp,  R1.shape, R11.shape)
            #             print()

            R2 = -np.matmul(Q.T, B)
            R21 = R2[:(Mz + 1), :]
            R22 = R2[(Mz + 1):, :]

            U, S, V = np.linalg.svd(R22, full_matrices=True)
            b = V.conj().T[:, -1]
            a = np.matmul(R21, b)
            a = -np.matmul(np.linalg.inv(R11), a)

            g0 = np.abs(a[-1] / b[-1]) if phys else a[-1] / b[-1]
            pa = np.polynomial.Polynomial(a)
            pb = np.polynomial.Polynomial(b)
            z = pa.roots()
            p = pb.roots()

            if (z.shape[0] > 0) and (p.shape[0] > 0):
                zIm = z[np.abs(np.real(z)) < dWorigin]
                zPairs = z[np.abs(np.real(z)) >= dWorigin]
                if zPairs.shape[0] > 0:
                    zPairs = zPairs[np.real(zPairs) > -dWorigin]

                pIm = p[np.abs(np.real(p)) < dWorigin]
                pPairs = p[np.abs(np.real(p)) >= dWorigin]
                if pPairs.shape[0] > 0:
                    pPairs = pPairs[np.real(pPairs) > -dWorigin]

                zeros = np.hstack([-np.flip(zPairs).conj(), 1j * np.imag(zIm), zPairs]) if phys else z
                poles = np.hstack([-np.flip(pPairs).conj(), 1j * np.imag(pIm), pPairs]) if phys else p
                Hwpred = SPM.G_SZF(W0, g0, poles, zeros)
                p, z = poles, zeros
                e = np.abs((Hwpred - H0) ** 2).mean() / e0 * 100
                if stability:
                    e += qStability * (np.imag(p) > 0).sum()

            error.append(e)
            g0n.append(g0)
            zn.append(z)
            pn.append(p)

    idMin = np.argmin(error)
    g0 = g0n[idMin]
    zeros = zn[idMin]
    poles = pn[idMin]

    return g0, poles, zeros


def ChainCauchyMethod(H0, W0, nmbWindows=10, nmbMaxPoles=40, dWorigin=1e-4, phys=True, stability=False,
                      qStability=.1, useLogPrec=False, logPrec=15, diffZPMax=5, splitEvenly=True):
    H0n = []
    W0n = []

    if splitEvenly:
        wSplit = np.array_split(W0, nmbWindows)
        for i in range(nmbWindows):
            w1 = wSplit[i][0]
            w2 = wSplit[i][-1]
            W0n.append(W0[(w1 <= W0) & (W0 <= w2)])
            H0n.append(H0[(w1 <= W0) & (W0 <= w2)])
    else:
        dW = (W0[-1] - W0[0]) / nmbWindows
        W0n = []
        H0n = []
        for i in range(nmbWindows):
            w1 = W0[0] + i * dW
            w2 = W0[0] + (i + 1) * dW if i < (nmbWindows - 1) else W0[-1]
            W0n.append(W0[(w1 <= W0) & (W0 <= w2)])
            H0n.append(H0[(w1 <= W0) & (W0 <= w2)])

    g0n = []
    pn = []
    zn = []
    for i in range(nmbWindows):
        W1 = W0n[i]
        H1 = H0n[i]
        g0, poles, zeros = CauchyMethodOptZP(H1, W1, nmbMaxPoles=nmbMaxPoles, dWorigin=dWorigin,
                                             diffZPMax=diffZPMax, phys=phys,
                                             stability=stability, qStability=qStability,
                                             useLogPrec=useLogPrec, logPrec=logPrec, plotSingularValues=False)
        g0n.append(g0)
        pn.append(poles)
        zn.append(zeros)

    return g0n, pn, zn, W0n, H0n


def StackChainResult(g0n, pn, zn, distMax=30.):
    nmbWindows = len(g0n)

    g0 = np.mean(g0n)
    p = np.array([])
    z = np.array([])
    for i in range(nmbWindows):
        pi = pn[i]
        zi = zn[i]
        p = np.hstack([p, pi])
        z = np.hstack([z, zi])

    p = p[np.abs(p) < distMax]
    z = z[np.abs(z) < distMax]

    # p = p[np.real(p) >= 0]
    # z = z[np.real(z) >= 0]

    return g0, p, z


def HardMergeChain(g0n, pn, zn, wn):
    hnr, poles, residues = 0., np.array([]), np.array([])
    for i in range(len(g0n)):
        w = wn[i]
        g0 = g0n[i]
        p = pn[i]
        z = zn[i]
        r = GetResidueFromCauchy(g0, p, z)
        h0 = GetHnrFromCauchy(g0, p, z, r, .2j)

        if i == 0:
            criterion = (np.real(p) <= w[-1])
            invC = np.logical_not(criterion)
            rC = r[criterion]
            pC = p[criterion]
            h0 += (r[invC] / p[invC]).sum()
        elif 0 < i < (len(g0n) - 1):
            criterion = (w[0] <= np.real(p)) & (np.real(p) <= w[-1])
            invC = np.logical_not(criterion)
            rC = r[criterion]
            pC = p[criterion]
            h0 += (r[invC] / p[invC]).sum()
        else:
            criterion = (w[0] <= np.real(p))
            invC = np.logical_not(criterion)
            rC = r[criterion]
            pC = p[criterion]
            h0 += (r[invC] / p[invC]).sum()

        hnr += h0
        poles = np.hstack([poles, pC])
        residues = np.hstack([residues, rC])

    return hnr / len(wn), poles, residues


def GetResidueFromCauchy(g0, poles, zeros):
    rp = np.zeros_like(poles) + 0j
    for i in range(poles.shape[0]):
        rp[i] = SPM.G_SZF(poles[i], g0, np.hstack([poles[:i], poles[(i + 1):]]), zeros)

    return rp


def GetHnrFromCauchy(g0, poles, zeros, residues, wa=0.):
    HNR = SPM.G_SZF(wa, g0, poles, zeros)
    HNR -= (residues / (wa - poles + 1e-10)).sum()

    return HNR


import torch
import torch.optim as optim

import numpy as np
from tqdm import trange

from . import ParameterManager as SPP



c = 3e8
hb = 6.582119570e-1



def H_terms(w, parameterList):
    v = []
    w1 = w[..., None]

    fctType = parameterList[0]
    tensorized = parameterList[1]
    stableEffectivePoles = parameterList[-1]

    if fctType == "sem":
        Hnr, pnC_R, pnC_I, rnC_R, rnC_I, pnI_I, rnI_I, pnE_R, pnE_I, rnE_R, rnE_I, r0, poleOrigin = parameterList[2:-1]

        v.append(Hnr)

        if tensorized:
            pnC = torch.abs(pnC_R) - 1j*torch.abs(pnC_I)
            rnC = rnC_R + 1j*rnC_I

            pnI = -1j * torch.abs(pnI_I)
            rnI = 1j * rnI_I/2

            pnE = torch.abs(pnE_R) - 1j*torch.abs(pnE_I) if stableEffectivePoles else torch.abs(pnE_R) + 1j*pnE_I
            rnE = rnE_R + 1j * rnE_I

            if poleOrigin:
                v.append(-1j*torch.abs(r0)/w1)

            v.append(rnC/(w1-pnC) - rnC.conj()/(w1+pnC.conj()))
            v.append(rnI/(w1-pnI) - rnI.conj()/(w1+pnI.conj()))
            v.append(rnE/(w1-pnE) - rnE.conj()/(w1+pnE.conj()))

        else:
            pnC = np.abs(pnC_R) - 1j * np.abs(pnC_I)
            rnC = rnC_R + 1j * rnC_I

            pnI = -1j * np.abs(pnI_I)
            rnI = 1j * rnI_I / 2

            pnE = np.abs(pnE_R) - 1j * np.abs(pnE_I) if stableEffectivePoles else np.abs(pnE_R) + 1j * pnE_I
            rnE = rnE_R + 1j * rnE_I

            if poleOrigin:
                v.append(-1j*np.abs(r0)/w1)

            v.append(rnC/(w1-pnC) - rnC.conj()/(w1+pnC.conj()))
            v.append(rnI/(w1-pnI) - rnI.conj()/(w1+pnI.conj()))
            v.append(rnE/(w1-pnE) - rnE.conj()/(w1+pnE.conj()))


    elif fctType == "gdl":
        Hnr, G0, an, bn, s1n, Gn, wn, s2n, pnE_R, pnE_I, rnE_R, rnE_I = parameterList[2:-1]

        v.append(1+Hnr)

        if tensorized:
            s1nn = s1n
            Gnn = 0.00005 + torch.abs(Gn)
            wnn = 0.00005 + torch.abs(wn)
            s2nn = s2n

            pnE = torch.abs(pnE_R) - 1j*torch.abs(pnE_I) if stableEffectivePoles else torch.abs(pnE_R) + 1j*pnE_I
            rnE = rnE_R + 1j * rnE_I

            v.append(1j*G0/w)
            v.append(-an**2/(w1**2 + 1j*torch.abs(bn)*w1))
            v.append(-(1j*s1nn*Gnn*w1 + s2nn*((Gnn/2)**2 + wnn**2))/((w1**2 - (Gnn/2)**2 - wnn**2) + 1j*Gnn*w1))
            v.append(rnE/(w1-pnE) - rnE.conj()/(w1+pnE.conj()))

        else:
            s1nn = s1n
            Gnn = 0.00005 + np.abs(Gn)
            wnn = 0.00005 + np.abs(wn)
            s2nn = s2n

            pnE = np.abs(pnE_R) - 1j*np.abs(pnE_I) if stableEffectivePoles else np.abs(pnE_R) + 1j*pnE_I
            rnE = rnE_R + 1j * rnE_I

            v.append(1j*G0/w)
            v.append(-an**2/(w1**2 + 1j*np.abs(bn)*w1))
            v.append(-(1j*s1nn*Gnn*w1 + s2nn*((Gnn/2)**2 + wnn**2))/((w1**2 - (Gnn/2)**2 - wnn**2) + 1j*Gnn*w1))
            v.append(rnE/(w1-pnE) - rnE.conj()/(w1+pnE.conj()))


    elif fctType == "mtse":
        Hnr, mrn, trn, wnR, wnI, pnE_R, pnE_I, rnE_R, rnE_I = parameterList[2:-1]

        v.append(Hnr)

        if tensorized:
            rn = torch.abs(mrn) * torch.exp(1j * trn)
            wn = torch.abs(wnR) - 1j * torch.abs(wnI)

            Cn = -rn / (2j * torch.imag(wn))
            Tn1 = 1j * torch.log((w1 - wn) / (w1 - wn.conj())) + trn + torch.pi/2
            Tn2 = 1j * torch.log((w1 + wn.conj()) / (w1 + wn)) - trn - torch.pi/2

            pnE = torch.abs(pnE_R) - 1j*torch.abs(pnE_I) if stableEffectivePoles else torch.abs(pnE_R) + 1j*pnE_I
            rnE = rnE_R + 1j * rnE_I

            v.append(2 * torch.real(Cn) + torch.abs(Cn) * (torch.exp(1j * Tn1) + torch.exp(1j * Tn2)))
            v.append(rnE/(w1-pnE) - rnE.conj()/(w1+pnE.conj()))

        else:
            rn = np.abs(mrn) * np.exp(1j * trn)
            wn = np.abs(wnR) + 1j * wnI

            Cn = -rn / (2j * np.imag(wn))
            theta = trn - np.pi / 2 + np.pi * (wnI > 0)
            Tn1 = 1j * np.log((w1 - wn) / (w1 - wn.conj())) + theta + np.pi
            Tn2 = 1j * np.log((w1 + wn.conj()) / (w1 + wn)) - theta + np.pi

            pnE = np.abs(pnE_R) - 1j*np.abs(pnE_I) if stableEffectivePoles else np.abs(pnE_R) + 1j*pnE_I
            rnE = rnE_R + 1j * rnE_I

            v.append(2 * np.real(Cn) + np.abs(Cn) * (np.exp(1j * Tn1) + np.exp(1j * Tn2)))
            v.append(rnE/(w1-pnE) - rnE.conj()/(w1+pnE.conj()))


    elif fctType == "szf":
        (H0, pnC_R, pnC_I, pnI_I, pnE_R, pnE_I, znC_R, znC_I, znI_I, znE_R, znE_I,
         poleOrigin, reversibleZeros) = parameterList[2:-1]

        v.append(H0)

        if tensorized:

            if poleOrigin:
                v.append(1/w1 + 0j)

            pnC = torch.abs(pnC_R) - 1j*torch.abs(pnC_I)
            znC = torch.abs(znC_R) + 1j*znC_I if not reversibleZeros else torch.abs(znC_R) - 1j*torch.abs(znC_I)
            v.append((w1-znC))
            v.append(1/(w1-pnC))
            v.append((w1+znC.conj()))
            v.append(1/(w1+pnC.conj()))

            pnI = -1j * torch.abs(pnI_I)
            znI = 1j * znI_I if (not reversibleZeros) else -1j * torch.abs(znI_I)
            v.append(w1-znI)
            v.append(1/(w1-pnI))

            pnE = torch.abs(pnE_R) - 1j*torch.abs(pnE_I) if stableEffectivePoles else torch.abs(pnE_R) + 1j*pnE_I
            znE = torch.abs(znE_R) - 1j*torch.abs(znE_I) if reversibleZeros else torch.abs(znE_R) + 1j*znE_I
            v.append((w1-znE))
            v.append(1/(w1-pnE))
            v.append((w1+znE.conj()))
            v.append(1/(w1+pnE.conj()))

        else:

            if poleOrigin:
                v.append(1/w1 + 0j)

            pnC = np.abs(pnC_R) - 1j*np.abs(pnC_I)
            znC = np.abs(znC_R) + 1j*znC_I if not reversibleZeros else np.abs(znC_R) - 1j*np.abs(znC_I)
            v.append((w1-znC))
            v.append(1/(w1-pnC))
            v.append((w1+znC.conj()))
            v.append(1/(w1+pnC.conj()))

            pnI = -1j * np.abs(pnI_I)
            znI = 1j * znI_I if (not reversibleZeros) else -1j * np.abs(znI_I)
            v.append(w1-znI)
            v.append(1/(w1-pnI))

            pnE = np.abs(pnE_R) - 1j*np.abs(pnE_I) if stableEffectivePoles else np.abs(pnE_R) + 1j*pnE_I
            znE = np.abs(znE_R) - 1j*np.abs(znE_I) if reversibleZeros else np.abs(znE_R) + 1j*znE_I
            v.append((w1-znE))
            v.append(1/(w1-pnE))
            v.append((w1+znE.conj()))
            v.append(1/(w1+pnE.conj()))

    else:
        print('Invalid fct type. Available options are "szf", "sem", "gdl" and "mtse" ')
        exit()

    return fctType, v



def H(fctType, H_terms, tensorized=True):
    v = 0.
    if fctType == "szf":
        v = 1.
        H0 = torch.real(H_terms[0]) if tensorized else np.real(H_terms[0])
        for resTerms in H_terms[1:]:
            if resTerms.ndim > 1:
                if tensorized:
                    v *= torch.prod(resTerms, dim=resTerms.ndim-1)
                else:
                    v *= np.prod(resTerms, axis=resTerms.ndim-1)
            else:
                v *= resTerms
        v *= H0
    else:
        Hnr = torch.real(H_terms[0]) if tensorized else np.real(H_terms[0])
        for resTerms in H_terms[1:]:
            if resTerms.ndim > 1:
                if tensorized:
                    v += resTerms.sum(dim=resTerms.ndim-1)
                else:
                    v += resTerms.sum(axis=resTerms.ndim-1)
            else:
                v += resTerms
        v += Hnr
    return v



def LossFct(Pr, Ta, alpha=(1,0,0,0), imaginaryPartPositive = False):

    l1 = torch.sum(torch.abs(Ta - Pr)**2) / torch.sum(torch.abs(Ta)**2) if alpha[0]>1e-5 else 0.
    l2 = torch.max(torch.abs(Ta - Pr) / (torch.abs(Ta) + 1e-3)) if alpha[1]>1e-5 else 0.

    dA = torch.abs(torch.real(Ta)) + 0.5 if alpha[2]>1e-5 else 0.
    A = torch.abs(torch.real(Ta-Pr)) if alpha[2]>1e-5 else 0.
    l3 = torch.mean(A/dA) if alpha[2]>1e-5 else 0.

    dB = torch.abs(torch.imag(Ta)) + 0.5 if alpha[3]>1e-5 else 0.
    B = torch.abs(torch.imag(Ta-Pr)) if alpha[3]>1e-5 else 0.
    l4 = torch.mean(B/dB) if alpha[3]>1e-5 else 0.

    l = alpha[0]*l1 + alpha[1]*l2 + alpha[2]*l3 + alpha[3]*l4

    if imaginaryPartPositive:
        l += 0.02*torch.relu( -torch.imag(Pr).sum() )

    return l




def FitModel(W, Ta, parameterList, imaginaryPartPositive=False, alpha=(1,0,0,0), tensorizeInput=False,
             nIter=2500, lr=3.3e-2, scIter=1200, scq=0.87,
             vizIteration=1000, visualizeLoss=True, loss_viz=None, logScalePoles=False):

    if tensorizeInput:
        W = torch.from_numpy(W)
        Ta = torch.from_numpy(Ta)

    fctType = parameterList[0]
    useZeros = True if fctType=="szf" else False
    scatterColorsZeros = None

    if useZeros:
        nmbPoles = 0
        poles, zeros = SPP.GetPolesAndZeros(parameterList)
        scatterColorsPoles = np.random.randint(0, 255, (poles.shape[0], 3))
        scatterColorsZeros = np.random.randint(0, 255, (zeros.shape[0], 3))

    else:
        nmbPoles = 0
        poles, residues = SPP.GetPolesAndResidues(parameterList)
        scatterColorsPoles = np.random.randint(0, 255, (poles.shape[0], 3))


    def lambda1(epoch):
        v = scq ** (epoch // scIter)
        # if lr * v < 5e-5:
        #     v = 5e-5 / lr
        return v

    optimizer = None
    if fctType == "sem":
        optimizer = optim.AdamW(parameterList[2:-2], lr=lr)
    elif fctType == "szf":
        optimizer = optim.AdamW(parameterList[2:-3], lr=lr)
    else:
        optimizer = optim.AdamW(parameterList[2:-1], lr=lr)
    scheduler = optim.lr_scheduler.LambdaLR(optimizer, lr_lambda=lambda1)

    lossValues = np.zeros(nIter)
    bar = trange(nIter, desc="Fitting SEM")
    for i in bar:

        optimizer.zero_grad()

        fctType, resonantTerms = H_terms(W, parameterList)
        Pr = H(fctType, resonantTerms, tensorized=True)

        Ln = LossFct(Pr, Ta, alpha=alpha, imaginaryPartPositive=imaginaryPartPositive)
        lossValues[i] = Ln.detach().numpy()

        if not (torch.isnan(Ln)):

            Ln.backward()
            optimizer.step()
            scheduler.step()

            i += 1

            if visualizeLoss:
                with torch.no_grad():
                    if (i % vizIteration) == 0:
                        # LOSS FUNCTION =============================================================================
                        # ===========================================================================================
                        vizDict = dict(xlabel="iteration",
                                       ylabel="loss",
                                       title="Cost - " + str(scheduler.get_last_lr()),
                                       ytype="log")
                        loss_viz.line(X=np.arange(i),
                                      Y=lossValues[:i],
                                      opts=vizDict,
                                      win="cost")



                        # Real, Imag, Abs curves ====================================================================
                        # ===========================================================================================
                        Ta_n = Ta.detach().numpy()
                        Pr_n = Pr.detach().numpy()


                        y = np.vstack([np.abs(Ta_n), np.abs(Pr_n)])
                        x = np.vstack([W, W])
                        vizDict = dict(xlabel="w",
                                       title="|H(w)|",
                                       xtype="log",
                                       legend=["Target", "Reconstructed"])
                        loss_viz.line(X=x.transpose(),
                                      Y=y.transpose(),
                                      opts=vizDict,
                                      win="abs H")

                        y = np.vstack([np.real(Ta_n), np.real(Pr_n)])
                        x = np.vstack([W, W])
                        vizDict = dict(xlabel="w",
                                       title="Re[H(w)]",
                                       xtype="log",
                                       legend=["Target", "Reconstructed"])
                        loss_viz.line(X=x.transpose(),
                                      Y=y.transpose(),
                                      opts=vizDict,
                                      win="real H")

                        y = np.vstack([np.imag(Ta_n), np.imag(Pr_n)])
                        x = np.vstack([W, W])
                        vizDict = dict(xlabel="w",
                                       title="Im[H(w)]",
                                       xtype="log",
                                       legend=["Target", "Reconstructed"])
                        loss_viz.line(X=x.transpose(),
                                      Y=y.transpose(),
                                      opts=vizDict,
                                      win="imag H")




                        # Poles, residues, zeros ====================================================================
                        # ===========================================================================================
                        polesArray, residuesArray, zerosArray = np.array([]), np.array([]), np.array([])
                        if useZeros:
                            poles, zeros = SPP.GetPolesAndZeros(parameterList)
                            polesArray = poles + 1e-2
                            zerosArray = zeros + 1e-2
                        else:
                            poles, residues = SPP.GetPolesAndResidues(parameterList)
                            polesArray = poles + 1e-2
                            residuesArray = residues


                        xtype = "log" if logScalePoles else "linear"
                        x = np.vstack([np.real(polesArray), np.imag(polesArray)]).T
                        vizDict = dict(xlabel="R",
                                       ylabel="iR",
                                       title="poles",
                                       xtype=xtype,
                                       markercolor=scatterColorsPoles)
                        loss_viz.scatter(X=x.transpose(0, 1),
                                         opts=vizDict,
                                         win="poles")

                        if useZeros:
                            x = np.vstack([np.real(zerosArray), np.imag(zerosArray)]).T
                            vizDict = dict(xlabel="R",
                                           ylabel="iR",
                                           title="zeros",
                                           xtype=xtype,
                                           markercolor=scatterColorsZeros)
                            loss_viz.scatter(X=x.transpose(0, 1),
                                             opts=vizDict,
                                             win="zeros")
                        else:
                            x = np.vstack([np.real(residuesArray), np.imag(residuesArray)]).T
                            vizDict = dict(xlabel="R",
                                           ylabel="iR",
                                           title="residues",
                                           markercolor=scatterColorsPoles)
                            loss_viz.scatter(X=x.transpose(0, 1),
                                             opts=vizDict,
                                             win="residues")

        else:
            print("ERREUR")
            break

    return parameterList



import matplotlib
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap


def GetAxCoordinates(ax):
    fig = ax.figure
    figH = fig.bbox.height
    figW = fig.bbox.width

    bbox = np.array([ax.bbox.x0 / figW, ax.bbox.y0 / figH, ax.bbox.width / figW, ax.bbox.height / figH])

    return bbox


def AmplitudeMap(Hw, W, ax=None, figID=1, fs=(5, 4), adjustFigure=False):
    cMap = np.array(sns.color_palette("Spectral_r", 15).as_hex())
    cMap = ListedColormap(cMap)

    cbar_ax_coord = None
    if ax is None:
        fig = plt.figure(figID, figsize=fs)
        plt.clf()
        ax = fig.subplots(1)
        ax_coord = GetAxCoordinates(ax)
        axRight = ax_coord[0] + ax_coord[2]
        if adjustFigure:
            axRight = 0.83
            fig.subplots_adjust(left=0.17, right=axRight, bottom=0.17, top=0.99)
        cbar_ax_coord = GetAxCoordinates(ax)
        cbar_ax_coord[0] = axRight + 0.013
        cbar_ax_coord[2] = min(0.04, 1 - axRight - 0.05)
    else:
        fig = ax.figure
        ax_coord = GetAxCoordinates(ax)
        axRight = ax_coord[0] + ax_coord[2]
        if adjustFigure:
            axRight = 0.83
            fig.subplots_adjust(left=0.17, right=axRight, bottom=0.17, top=0.99)
        cbar_ax_coord = GetAxCoordinates(ax)
        cbar_ax_coord[0] = axRight + 0.013
        cbar_ax_coord[2] = min(0.04, 1 - axRight - 0.05)

    a = ax.imshow(np.real(np.log(Hw)), extent=[np.min(np.real(W)), np.max(np.real(W)),
                                               np.min(np.imag(W)), np.max(np.imag(W))],
                  origin="lower", aspect="auto", cmap=cMap)
    cbar_ax = fig.add_axes(cbar_ax_coord)
    fig.colorbar(a, cax=cbar_ax, orientation='vertical')

    ax.set_xlabel(r"Re[$\omega$] ($10^{15}$ rad/s)", fontsize=11)
    ax.set_xlim(np.min(np.real(W)), np.max(np.real(W)))
    ax.set_ylabel(r"Im[$\omega$] ($10^{15}$ rad/s)", fontsize=11)
    ax.set_ylim(np.min(np.imag(W)[:, 0]), np.max(np.imag(W)[:, 0]))

    return ax.figure, ax


def AmplitudeCurve(Hw, W, ax=None, figID=1, fs=(5, 4)):
    if ax is None:
        fig = plt.figure(figID, figsize=fs)
        plt.clf()
        ax = fig.subplots(1)
    else:
        fig = ax.figure

    ax.plot(W, np.abs(Hw), color="black")
    ax.set_xlabel(r"$\omega$ ($10^{15}$ rad/s)", fontsize=11)
    ax.set_xlim(W[0], W[-1])
    ax.set_ylabel("Amplitude", fontsize=11)

    return ax.figure, ax


def PhaseMap(Hw, W, ax=None, figID=1, fs=(5, 4), adjustFigure=False):
    cMap = np.array(sns.color_palette("YlGnBu_r", 15).as_hex())
    cMap = ListedColormap(cMap)

    cbar_ax_coord = None
    if ax is None:
        fig = plt.figure(figID, figsize=fs)
        plt.clf()
        ax = fig.subplots(1)
        ax_coord = GetAxCoordinates(ax)
        axRight = ax_coord[0] + ax_coord[2]
        if adjustFigure:
            axRight = 0.83
            fig.subplots_adjust(left=0.16, right=axRight, bottom=0.15, top=0.97)
        cbar_ax_coord = GetAxCoordinates(ax)
        cbar_ax_coord[0] = axRight + 0.013
        cbar_ax_coord[2] = min(0.04, 1 - axRight - 0.05)
    else:
        fig = ax.figure
        ax_coord = GetAxCoordinates(ax)
        axRight = ax_coord[0] + ax_coord[2]
        if adjustFigure:
            axRight = 0.83
            fig.subplots_adjust(left=0.16, right=axRight, bottom=0.15, top=0.97)
        cbar_ax_coord = GetAxCoordinates(ax)
        cbar_ax_coord[0] = axRight + 0.013
        cbar_ax_coord[2] = min(0.04, 1 - axRight - 0.05)

    a = ax.imshow(np.imag(np.log(Hw)) * 180 / np.pi, extent=[np.min(np.real(W)), np.max(np.real(W)),
                                                             np.min(np.imag(W)), np.max(np.imag(W))],
                  origin="lower", aspect="auto", cmap=cMap)
    cbar_ax = fig.add_axes(cbar_ax_coord)
    fig.colorbar(a, cax=cbar_ax, orientation='vertical')

    ax.set_xlabel(r"Re[$\omega$] ($10^{15}$ rad/s)", fontsize=11)
    ax.set_xlim(np.min(np.real(W)[0, :]), np.max(np.real(W)[0, :]))
    ax.set_ylabel(r"Im[$\omega$] ($10^{15}$ rad/s)", fontsize=11)
    ax.set_ylim(np.min(np.imag(W)[:, 0]), np.max(np.imag(W)[:, 0]))

    return ax.figure, ax


def PhaseCurve(Hw, W, ax=None, figID=1, fs=(5, 4)):
    if ax is None:
        fig = plt.figure(figID, figsize=fs)
        plt.clf()
        ax = fig.subplots(1)
    else:
        fig = ax.figure

    ax.plot(W, np.unwrap(np.imag(np.log(Hw))) * 180 / np.pi, color="black")
    ax.set_xlabel(r"$\omega$ ($10^{15}$ rad/s)", fontsize=11)
    ax.set_xlim(W[0], W[-1])
    ax.set_ylabel(r"Phase ($\degree$)", fontsize=11)

    return ax.figure, ax


def BodeDiagram(Hw, W, figID=1, fs=(5, 4)):
    fig = plt.figure(figID, figsize=fs)
    plt.clf()
    axs = fig.subplots(2, 1)

    fig, ax1 = AmplitudeCurve(Hw, W, ax=axs[0])
    ax1.set_xlabel("")
    fig, ax2 = PhaseCurve(Hw, W, ax=axs[1])

    return fig


def BodeMaps(Hw, W, figID=1, fs=(5, 4)):
    fig = plt.figure(figID, figsize=fs)
    plt.clf()
    axs = fig.subplots(2, 1)

    fig, ax1 = AmplitudeMap(Hw, W, ax=axs[0], adjustFigure=True)
    ax1.set_xlabel("")
    fig, ax2 = PhaseMap(Hw, W, ax=axs[1], adjustFigure=False)

    return fig


def AmplitudeFig(Hx, Wx, Hw, W, figID=1, fs=(5, 4)):
    fig = plt.figure(figID, figsize=fs)
    plt.clf()
    axs = fig.subplots(2, 1)

    fig, ax1 = AmplitudeCurve(Hw, W, ax=axs[0])
    ax1.set_xlabel("")
    ax1.set_xlim(min(np.min(np.real(Wx)), W[0]), min(np.max(np.real(Wx)), W[-1]))
    fig, ax2 = AmplitudeMap(Hx, Wx, ax=axs[1], adjustFigure=True)
    ax2.set_xlim(min(np.min(np.real(Wx)), W[0]), min(np.max(np.real(Wx)), W[-1]))

    return fig


def PhaseFig(Hx, Wx, Hw, W, figID=1, fs=(5, 4)):
    fig = plt.figure(figID, figsize=fs)
    plt.clf()
    axs = fig.subplots(2, 1)

    fig, ax1 = PhaseCurve(Hw, W, ax=axs[0])
    ax1.set_xlabel("")
    ax1.set_xlim(min(np.min(np.real(Wx)), W[0]), min(np.max(np.real(Wx)), W[-1]))
    fig, ax2 = PhaseMap(Hx, Wx, ax=axs[1], adjustFigure=True)
    ax2.set_xlim(min(np.min(np.real(Wx)), W[0]), min(np.max(np.real(Wx)), W[-1]))

    return fig


def ParameterDistribution(poles, figID=1, fs=(5, 4), wr1=None, wr2=None, wi1=None, wi2=None, freq=True):
    fig = plt.figure(figID, figsize=fs)
    plt.clf()
    ax = fig.subplots(1)
    fig.subplots_adjust(left=0.16, bottom=0.15)

    ax.scatter(x=np.real(poles), y=np.imag(poles), s=12, color="red", zorder=200)

    if freq:
        ax.set_xlabel(r"Re[$\omega$] ($10^{15}$ rad/s)", fontsize=11)
        ax.set_ylabel(r"Im[$\omega$] ($10^{15}$ rad/s)", fontsize=11)

    wr1 = np.min(np.real(poles)) - 0.1 if wr1 is None else wr1
    wr2 = np.max(np.real(poles)) + 0.1 if wr2 is None else wr2
    wi1 = np.min(np.imag(poles)) - 0.1 if wi1 is None else wi1
    wi2 = np.max(np.imag(poles)) + 0.1 if wi2 is None else wi2

    ax.set_xlim(wr1, wr2)
    ax.set_ylim(wi1, wi2)

    return fig, ax


import numpy as np

c = 3e8
def G_ab(w, a, b):
    dP = a.shape[0] - 1
    dQ = b.shape[0] - 1
    w1 = w
    if np.isscalar(w):
        w1 = [w]
    va = np.polynomial.polynomial.polyvander(w1, dP)
    vb = np.polynomial.polynomial.polyvander(w1, dQ)
    p = (va * a).sum(axis=len(va.shape) - 1)
    q = (vb * b).sum(axis=len(vb.shape) - 1)

    return p / q


def G_SZF(w, g0, poles, zeros):
    w1 = w[..., None] if not np.isscalar(w) else np.array([w])[..., None]

    v = (g0 * np.prod(w1-zeros, axis=len(w1.shape)-1) /
         np.prod(w1-poles, axis=len(w1.shape)-1))
    return v


def H_SEM(w, H_NR, p, r):
    w1 = w[..., np.newaxis]
    v = H_NR + (r/(w1-p)).sum(axis=len(w1.shape)-1)
    return v


def H_HermitianSEM(w, H_NR, p, r):
    w1 = w[..., np.newaxis]
    v = H_NR + (r/(w1-p) - r.conj()/(w1+p.conj())).sum(axis=len(w1.shape)-1)
    # dK = 0
    # v = (-1)**dK * r/(w1 - p)**(dK+1) - (-1)**dK * r.conj()/(w1 + p.conj())**(dK+1)
    # v += H_NR + v.sum(axis=w.ndim)
    return v


def MergePoles(poles, residues, W, d0=0.5, nPts=120):
    p = poles.copy()
    r = residues.copy()
    wa = W[0]
    wb = W[-1]
    w = np.linspace(wa, wb, nPts)

    Hterms = (r/(w[..., None]-p) - r.conj()/(w[..., None]-p.conj()))
    keepMerging = True
    while keepMerging:
        D = np.ones((p.shape[0], p.shape[0])) * 1e10
        for a in range(p.shape[0] - 1):
            ra = r[a]
            Ra = np.abs(ra)
            pa = p[a]
            ha = Hterms[:, a]
            for b in range(a + 1, p.shape[0]):
                rb = r[b]
                Rb = np.abs(rb)
                pb = p[b]
                hb = Hterms[:, b]

                rc = ra + rb
                RC = Ra + Rb
                pc = 1 / RC * (Ra * pa + Rb * pb)
                hc = (rc / (w - pc) - rc.conj() / (w - pc.conj()))

                D[a, b] = np.abs(hc - (ha + hb)).sum() / (np.abs(ha + hb).sum()) * 100

        minDID = np.argmin(D)
        rID, cID = minDID // D.shape[1], minDID % D.shape[1]
        minD = D[rID, cID]
        if minD < d0:
            id1 = min(rID, cID)
            id2 = max(rID, cID)

            r1 = r[id1]
            R1 = np.abs(r1)
            p1 = p[id1]
            r2 = r[id2]
            R2 = np.abs(r2)
            p2 = p[id2]
            rc = r1 + r2
            RC = R1 + R2
            pc = 1 / RC * (R1 * p1 + R2 * p2)

            p = np.hstack([p[:id1], p[(id1 + 1):id2], p[(id2 + 1):], pc])
            r = np.hstack([r[:id1], r[(id1 + 1):id2], r[(id2 + 1):], rc[..., None]])
        else:
            keepMerging = False

    return p,r


def RemovePoles(hnr, poles, residues, W, d0=0.1, s0=1, nPts=120):
    p = poles.copy()
    r = residues.copy()
    wa = W[0]
    wb = W[-1]
    w = np.linspace(wa, wb, nPts)

    Hterms = (r / (w[..., None] - p) - r.conj() / (w[..., None] - p.conj()))
    Yw = hnr + Hterms.sum(axis=1)

    idsToRemove = []
    for a in range(p.shape[0] - 1):
        ra = r[a]
        pa = p[a]
        ha = Hterms[:, a]

        s = np.std(np.abs(ha)) / np.std(np.abs(Yw)) * 100
        if s < s0:
            idsToRemove.append(a)
            hnr += 2 * np.real(ra / pa)
        else:
            d = np.abs(ha - Yw).sum() / (np.abs(Yw).sum()) * 100
            if d > (100 - d0):
                idsToRemove.append(a)

    if len(idsToRemove) > 0:
        p = np.delete(p, idsToRemove)
        r = np.delete(r, idsToRemove)

    return hnr, p, r


def StabilizePoles(p, r, q0=1e-5):
    rC = []
    pC = []
    for i in range(p.shape[0]):
        pi = p[i]
        ri = r[i]
        if np.abs(np.imag(pi)) < 1e-10:
            pC.append(np.real(pi) - 1j * q0)
            rC.append(ri / 2)
            pC.append(np.real(pi) + 1j * q0)
            rC.append(ri.conj() / 2)
        else:
            pC.append(pi)
            rC.append(ri)
    pC = np.array(pC)
    rC = np.array(rC)

    return pC, rC


import torch
import numpy as np

from . import Cauchy as SPC

c = 3e8
hb = 6.582119570e-1


# hb = 1


def Tensorize(parameterList):
    if not parameterList[1]:
        fctType = parameterList[0]

        paramRange = np.arange(2, len(parameterList)-1)
        if fctType=="sem":
            paramRange = np.arange(2, len(parameterList)-2)
        elif fctType=="szf":
            paramRange = np.arange(2, len(parameterList)-3)

        for i in paramRange:
            p = parameterList[i]
            if np.isscalar(p):
                parameterList[i] = torch.tensor(p)
            else:
                parameterList[i] = torch.from_numpy(p)
            parameterList[i].requires_grad_(True)

        parameterList[1] = True

    return parameterList



def Numpify(oldParameterList):
    parameterList = None
    if oldParameterList[1]:
        parameterList = [oldParameterList[0], oldParameterList[1]]
        fctType = oldParameterList[0]

        paramRange = np.arange(2, len(oldParameterList) - 1)
        if fctType == "sem":
            paramRange = np.arange(2, len(oldParameterList) - 2)
        elif fctType == "szf":
            paramRange = np.arange(2, len(oldParameterList) - 3)

        for i in paramRange:
            p = oldParameterList[i].clone()
            if torch.is_tensor(p):
                parameterList.append(p.detach().cpu().numpy())

        parameterList += oldParameterList[paramRange[-1]+1:]
        parameterList[1] = False

    else:
        parameterList = oldParameterList

    return parameterList



def GenerateSEMParameterList(fctType="sem", tensorized=True,
                             Hnr=None, residues=None, poles=None,
                             nPoles=0, nPolesIm=0, nPolesEff=0, stableEffectivePoles=False, poleOrigin=False,
                             w=None, minDist=2e-4):
    parameterList = None

    Hnr = np.random.randn(1) if (Hnr is None) else Hnr

    polesAdd = None
    residuesAdd = None
    if w is None:
        polesAdd = [np.abs(np.random.randn(nPoles) * 5) - 1j * np.abs(np.random.randn(nPoles) * 5)]
        polesAdd.append(-1j * np.abs(np.random.randn(nPolesIm)) * 0.001)
        if stableEffectivePoles:
            polesAdd.append(np.abs(np.random.randn(nPolesEff) * 5) - 1j * np.abs(np.random.randn(nPolesEff) * 5))
        else:
            polesAdd.append(np.abs(np.random.randn(nPolesEff) * 5) + 1j * np.random.randn(nPolesEff) * 5)
    else:
        wa = 0.7 * w[0]
        wb = 1.3 * w[-1]
        polesAdd = [np.abs(np.linspace(wa, wb, nPoles)) - 0.05j * np.abs(np.linspace(wa, wb, nPoles))]
        polesAdd.append(-0.01j * np.abs(np.linspace(wa, wb, nPolesIm)))
        if stableEffectivePoles:
            polesAdd.append(np.abs(np.linspace(1.2 * wb, 2 * wb, nPolesEff)) - 0.01j * np.abs(
                np.linspace(1.2 * wb, 2 * wb, nPolesEff)))
        else:
            polesAdd.append(
                np.abs(np.linspace(1.2 * wb, 2 * wb, nPolesEff)) + 0.01j * np.linspace(1.2 * wb, 2 * wb, nPolesEff))

    residuesAdd = [np.random.randn(nPoles) * 0.05 + 1j * np.random.randn(nPoles) * 0.05]
    residuesAdd.append(1j * np.random.randn(nPolesIm) * 5)
    residuesAdd.append(np.random.randn(nPolesEff) * 0.05 + 1j * np.random.randn(nPolesEff) * 0.05)

    if not poles is None:
        if residues is not None:
            residuesC = np.hstack([residues[np.real(poles) >= minDist], residuesAdd[0]])
            residuesI = np.hstack([residues[np.abs(np.real(poles)) < minDist], residuesAdd[1]])
            residuesEff = residuesAdd[2]
            residues = [residuesC, residuesI, residuesEff]

        polesC = np.hstack([poles[np.real(poles) >= minDist], polesAdd[0]])
        polesI = np.hstack([poles[np.abs(np.real(poles)) < minDist], polesAdd[1]])
        polesEff = polesAdd[2]
        poles = [polesC, polesI, polesEff]

        if residues is None:
            residues = [np.random.randn(polesC.shape[0]) * 5 + 1j * np.random.randn(polesC.shape[0]) * 5]
            residues.append(1j * np.random.randn(polesI.shape[0]) * 5)
            residues.append(np.random.randn(polesEff.shape[0]) * 5 + 1j * np.random.randn(polesEff.shape[0]) * 5)

    else:
        residues = [residuesAdd[0], residuesAdd[1], residuesAdd[2]]
        poles = [polesAdd[0], polesAdd[1], polesAdd[2]]

    if fctType == "sem":
        pnC, pnI, pnE = poles
        pnC_R, pnC_I = np.real(pnC), np.imag(pnC)
        pnI_I = np.imag(pnI)
        pnE_R, pnE_I = np.real(pnE), np.imag(pnE)

        rnC, rnI, rnE = residues
        r0 = 0.
        if poleOrigin:
            r0 = np.random.randn(1) * 5j
        rnC_R, rnC_I = np.real(rnC), np.imag(rnC)
        rnI_I = np.imag(rnI)
        rnE_R, rnE_I = np.real(rnE), np.imag(rnE)

        parameterList = [fctType, False, Hnr, pnC_R, pnC_I, rnC_R, rnC_I, pnI_I, rnI_I, pnE_R, pnE_I, rnE_R, rnE_I,
                         r0, poleOrigin, stableEffectivePoles]

    elif fctType == "gdl":
        pnC, pnI, pnE = poles
        pnC_R, pnC_I = np.real(pnC), np.imag(pnC)
        pnI_I = np.imag(pnI)
        pnE_R, pnE_I = np.real(pnE), np.imag(pnE)

        rnC, rnI, rnE = residues
        rnC_R, rnC_I = np.real(rnC), np.imag(rnC)
        rnE_R, rnE_I = np.real(rnE), np.imag(rnE)

        bn = -np.abs(pnI_I)
        an = np.sqrt(np.abs(bn * rnI))
        r0 = np.random.randn(1)
        G0 = r0 + ((an ** 2) / bn).sum()

        wn = np.abs(pnC_R)
        Gn = -2 * np.abs(pnC_I)
        s1n = -2 * rnC_I / Gn
        s2n = -2 * np.real(rnC * pnC.conj()) / (np.abs(pnC) ** 2)

        parameterList = [fctType, False, Hnr, G0, an, bn, s1n, Gn, wn, s2n, pnE_R, pnE_I, rnE_R, rnE_I,
                         stableEffectivePoles]

    elif fctType == "mtse":
        pnC, pnI, pnE = poles
        pnC_R, pnC_I = np.real(pnC), np.imag(pnC)
        pnI_I = np.imag(pnI)
        pnE_R, pnE_I = np.real(pnE), np.imag(pnE)

        rnC, rnI, rnE = residues
        rnE_R, rnE_I = np.real(rnE), np.imag(rnE)
        mrn = np.abs(rnC)
        trn = np.angle(rnC)

        parameterList = [fctType, False, Hnr, mrn, trn, pnC_R, pnC_I, pnE_R, pnE_I, rnE_R, rnE_I, stableEffectivePoles]

    else:
        print('Invalid fct type. Available options are "sem", "gdl" and "mtse" ')
        exit()

    if tensorized and (parameterList is not None):
        parameterList = Tensorize(parameterList)

    return parameterList


def GenerateSZFParameterList(tensorized=True, G0=None, poles=None, zeros=None,
                             nPoles=0, nPolesIm=0, nPolesEff=0, stableEffectivePoles=False, poleOrigin=False,
                             nZeros=0, nZerosIm=0, nZerosEff=0, reversibleZeros=False,
                             w=None, minDist=2e-4):
    parameterList = None

    G0 = np.random.randn(1) if (G0 is None) else G0

    polesAdd = None
    zerosAdd = None
    if w is None:
        polesAdd = [np.abs(np.random.randn(nPoles) * 5) - 1j * np.abs(np.random.randn(nPoles) * 5)]
        polesAdd.append(-1j * np.abs(np.random.randn(nPolesIm)) * 0.001)
        if stableEffectivePoles:
            polesAdd.append(np.abs(np.random.randn(nPolesEff) * 5) - 1j * np.abs(np.random.randn(nPolesEff) * 5))
        else:
            polesAdd.append(np.abs(np.random.randn(nPolesEff) * 5) + 1j * np.random.randn(nPolesEff) * 5)

        if reversibleZeros:
            zerosAdd = [np.abs(np.random.randn(nZeros) * 5) - 1j * np.abs(np.random.randn(nZeros) * 5)]
            zerosAdd.append(-1j * np.abs(np.random.randn(nZerosIm)) * 0.001)
            zerosAdd.append(np.abs(np.random.randn(nZerosEff) * 5) - 1j * np.abs(np.random.randn(nZerosEff)) * 5)
        else:
            zerosAdd = [np.abs(np.random.randn(nZeros) * 5) + 1j * np.random.randn(nZeros) * 5]
            zerosAdd.append(1j * np.random.randn(nZerosIm) * 0.001)
            zerosAdd.append(np.abs(np.random.randn(nZerosEff) * 5) + 1j * np.random.randn(nZerosEff) * 5)
    else:
        wa = 0.7 * w[0]
        wb = 1.3 * w[-1]
        polesAdd = [np.abs(np.linspace(wa, wb, nPoles)) - 0.1j * np.abs(np.linspace(wa, wb, nPoles))]
        polesAdd.append(-0.1j * np.abs(np.linspace(wa, wb, nPolesIm)))
        if stableEffectivePoles:
            polesAdd.append(np.abs(np.linspace(1.2 * wb, 2 * wb, nPolesEff)) - 0.1j * np.abs(
                np.linspace(1.2 * wb, 2 * wb, nPolesEff)))
        else:
            polesAdd.append(
                np.abs(np.linspace(1.2 * wb, 2 * wb, nPolesEff)) + 0.1j * np.linspace(1.2 * wb, 2 * wb, nPolesEff))

        if reversibleZeros:
            zerosAdd = [np.abs(np.linspace(wa, wb, nZeros)) - 0.2j * np.abs(np.linspace(wa, wb, nZeros))]
            zerosAdd.append(-0.2j * np.abs(np.linspace(wa, wb, nZerosIm)))
            zerosAdd.append(np.abs(np.linspace(1.2 * wb, 2 * wb, nZerosEff)) - 0.2j * np.abs(
                np.linspace(1.2 * wb, 2 * wb, nZerosEff)))
        else:
            zerosAdd = [np.abs(np.linspace(wa, wb, nZeros)) + 0.2j * np.linspace(wa, wb, nZeros)]
            zerosAdd.append(0.2j * np.linspace(wa, wb, nZerosIm))
            zerosAdd.append(
                np.abs(np.linspace(1.2 * wb, 2 * wb, nZerosEff)) + 0.2j * np.linspace(1.2 * wb, 2 * wb, nZerosEff))

    if not poles is None:
        polesC = np.hstack([poles[np.real(poles) >= minDist], polesAdd[0]])
        polesI = np.hstack([poles[np.abs(np.real(poles)) < minDist], polesAdd[1]])
        polesEff = polesAdd[2]
        poles = [polesC, polesI, polesEff]

    else:
        poles = [polesAdd[0], polesAdd[1], polesAdd[2]]

    if not zeros is None:
        zerosC = np.hstack([zeros[np.real(zeros) >= minDist], zerosAdd[0]])
        zerosI = np.hstack([zeros[np.abs(np.real(zeros)) < minDist], zerosAdd[1]])
        zerosEff = zerosAdd[2]
        zeros = [zerosC, zerosI, zerosEff]

    else:
        zeros = [zerosAdd[0], zerosAdd[1], zerosAdd[2]]

    pnC, pnI, pnE = poles
    pnC_R, pnC_I = np.real(pnC), np.imag(pnC)
    pnI_I = np.imag(pnI)
    pnE_R, pnE_I = np.real(pnE), np.imag(pnE)

    znC, znI, znE = zeros
    znC_R, znC_I = np.real(znC), np.imag(znC)
    znI_I = np.imag(znI)
    znE_R, znE_I = np.real(znE), np.imag(znE)

    parameterList = ["szf", False, G0, pnC_R, pnC_I, pnI_I, pnE_R, pnE_I, znC_R, znC_I, znI_I, znE_R, znE_I,
                     poleOrigin, reversibleZeros, stableEffectivePoles]

    if tensorized and (parameterList is not None):
        parameterList = Tensorize(parameterList)

    return parameterList


def ConvertToSEM(parameterList, wa=0.):
    newParameterList = None

    fctType = parameterList[0]
    tensorized = parameterList[1]
    stableEffectivePoles = parameterList[-1]

    if tensorized:
        parameterList = Numpify(parameterList)

    Hnr = parameterList[2]

    if fctType == "gdl":
        G0, an, bn, s1n, Gn, wn, s2n, pnE_R, pnE_I, rnE_R, rnE_I = parameterList[3:-1]

        s1nn = s1n
        Gnn = 0.00005 + np.abs(Gn)
        wnn = 0.00005 + np.abs(wn)
        s2nn = s2n

        pnE = np.abs(pnE_R) - 1j * np.abs(pnE_I) if stableEffectivePoles else np.abs(pnE_R) + 1j * pnE_I
        pnC = wnn - 1j * Gnn / 2
        pnI = -1j * np.abs(bn)

        rnE = rnE_R + 1j * rnE_I
        rnC = -(1j * s1nn * Gnn * (wnn - 1j * (Gnn / 2)) + s2nn * (wnn ** 2 + (Gnn / 2) ** 2)) / (2 * wnn)
        rnI = -1j * an ** 2 / np.abs(bn)

        r0 = 1j * G0 - rnI.sum()
        poleOrigin = True

        newParameterList = ["sem", tensorized, Hnr,
                            np.real(pnC), np.imag(pnC), np.real(rnC), np.imag(rnC),
                            np.imag(pnI), np.imag(rnI),
                            np.real(pnE), np.real(pnE), np.real(rnE), np.real(rnE),
                            r0, poleOrigin, stableEffectivePoles]


    elif fctType == "mtse":
        mrn, trn, pnC_R, pnC_I, pnE_R, pnE_I, rnE_R, rnE_I = parameterList[3:-1]

        rnC = np.abs(mrn) * np.exp(1j * trn)
        rnC_R, rnC_I = np.real(rnC), np.imag(rnC)
        r0 = 0.

        poleOrigin = False

        newParameterList = ["sem", tensorized, Hnr,
                            pnC_R, pnC_I, rnC_R, rnC_I,
                            np.array([]), np.array([]),
                            pnE_R, pnE_I, rnE_R, rnE_I,
                            r0, poleOrigin, stableEffectivePoles]


    elif fctType == "szf":
        (G0, pnC_R, pnC_I, pnI_I, pnE_R, pnE_I, znC_R, znC_I, znI_I, znE_R, znE_I,
         poleOrigin, reversibleZeros) = parameterList[2:-1]

        pnC = np.abs(pnC_R) - 1j * np.abs(pnC_I)
        znC = np.abs(znC_R) + 1j * znC_I if not reversibleZeros else np.abs(znC_R) - 1j * np.abs(znC_I)

        pnI = -1j * np.abs(pnI_I)
        znI = 1j * znI_I / 2 if (not reversibleZeros) else -1j * np.abs(znI_I) / 2

        pnE = np.abs(pnE_R) - 1j * np.abs(pnE_I) if stableEffectivePoles else np.abs(pnE_R) + 1j * pnE_I
        znE = np.abs(znE_R) - 1j * np.abs(znE_I) if reversibleZeros else np.abs(znE_R) + 1j * znE_I

        poles = np.hstack([pnC, pnI, pnE])
        zeros = np.hstack([znC, znI, znE])

        residues = SPC.GetResidueFromCauchy(G0, poles, zeros)
        HNR = SPC.GetHnrFromCauchy(G0, poles, zeros, residues, wa=wa)

        newParameterList = GenerateSEMParameterList(fctType="sem", tensorized=tensorized,
                                                    Hnr=HNR, residues=residues, poles=poles,
                                                    nPoles=0, nPolesIm=0, nPolesEff=0,
                                                    stableEffectivePoles=stableEffectivePoles, poleOrigin=poleOrigin)


    else:
        newParameterList = parameterList

    if tensorized:
        newParameterList = Tensorize(newParameterList)

    return newParameterList


def ConvertToGDL(oldParameterList, wa=0.):
    newParameterList = None

    fctType = oldParameterList[0]
    tensorized = oldParameterList[1]
    stableEffectivePoles = oldParameterList[-1]

    parameterList = None
    if not (fctType == "sem" or fctType == "gdl"):
        parameterList = ConvertToSEM(oldParameterList, wa=wa)
    else:
        parameterList = oldParameterList

    if tensorized:
        parameterList = Numpify(parameterList)

    fctType = parameterList[0]
    if fctType == "sem":
        Hnr, pnC_R, pnC_I, rnC_R, rnC_I, pnI_I, rnI_I, pnE_R, pnE_I, rnE_R, rnE_I, r0, poleOrigin = parameterList[2:-1]

        pnC = np.abs(pnC_R) - 1j * np.abs(pnC_I)
        rnC = rnC_R + 1j * rnC_I

        pnI = -1j * np.abs(pnI_I)
        rnI = 1j * rnI_I / 2

        pnE = np.abs(pnE_R) - 1j * np.abs(pnE_I) if stableEffectivePoles else np.abs(pnE_R) + 1j * pnE_I
        pnE_R, pnE_I = np.real(pnE), np.imag(pnE)
        rnE = rnE_R + 1j * rnE_I
        rnE_R, rnE_I = np.real(rnE), np.imag(rnE)

        bn = -np.abs(pnI)
        an = np.sqrt(np.abs(bn * rnI))
        r0 = np.random.randn(1)
        G0 = r0 + ((an ** 2) / bn).sum()

        wn = np.abs(pnC_R)
        Gn = -2 * np.abs(pnC_I)
        s1n = -2 * rnC_I / Gn
        s2n = -2 * np.real(rnC * pnC.conj()) / (np.abs(pnC) ** 2)

        newParameterList = ["gdl", tensorized, Hnr, G0, an, bn, s1n, Gn, wn, s2n,
                            pnE_R, pnE_I, rnE_R, rnE_I,
                            stableEffectivePoles]

    else:
        newParameterList = parameterList

    if tensorized:
        newParameterList[1] = False
        newParameterList = Tensorize(newParameterList)

    return newParameterList


def ConvertToMTSE(oldParameterList, wa=0.):
    newParameterList = None

    fctType = oldParameterList[0]
    tensorized = oldParameterList[1]
    stableEffectivePoles = oldParameterList[-1]

    parameterList = None
    if not (fctType == "sem" or fctType == "mtse"):
        parameterList = ConvertToSEM(oldParameterList, wa=wa)
    else:
        parameterList = oldParameterList

    if tensorized:
        parameterList = Numpify(parameterList)

    fctType = parameterList[0]
    if fctType == "sem":
        Hnr, pnC_R, pnC_I, rnC_R, rnC_I, pnI_I, rnI_I, pnE_R, pnE_I, rnE_R, rnE_I, r0, poleOrigin = parameterList[2:-1]

        pnC = np.abs(pnC_R) - 1j * np.abs(pnC_I)
        pnI = -1j * np.abs(pnI_I)
        pnE = np.abs(pnE_R) - 1j * np.abs(pnE_I) if stableEffectivePoles else np.abs(pnE_R) + 1j * pnE_I

        rnC = rnC_R + 1j * rnC_I
        rnI = 1j * rnI_I / 2
        rnE = rnE_R + 1j * rnE_I

        pnE = np.hstack([pnI, pnE])
        rnE = np.hstack([rnI, rnE])

        mrn = np.abs(rnC)
        trn = np.angle(rnC)

        newParameterList = ["mtse", tensorized, Hnr,
                            mrn, trn, np.real(pnC), np.imag(pnC),
                            np.real(pnE), np.imag(pnE), np.real(rnE), np.imag(rnE),
                            stableEffectivePoles]

    else:
        newParameterList = parameterList

    if tensorized:
        newParameterList[1] = False
        newParameterList = Tensorize(newParameterList)

    return newParameterList


def GetPolesAndResidues(oldParameterList, wa=0.):
    poles, residues = None, None

    fctType = oldParameterList[0]
    tensorized = oldParameterList[1]
    stableEffectivePoles = oldParameterList[-1]

    parameterList = None
    if fctType == "szf":
        parameterList = ConvertToSEM(oldParameterList, wa=wa)
    else:
        parameterList = oldParameterList if not tensorized else Numpify(oldParameterList)

    if tensorized:
        parameterList = Numpify(parameterList)

    fctType = oldParameterList[0]
    if fctType == "sem":
        Hnr, pnC_R, pnC_I, rnC_R, rnC_I, pnI_I, rnI_I, pnE_R, pnE_I, rnE_R, rnE_I, r0, poleOrigin = parameterList[2:-1]

        pnC = np.abs(pnC_R) - 1j * np.abs(pnC_I)
        rnC = rnC_R + 1j * rnC_I

        pnI = -1j * np.abs(pnI_I)
        rnI = 1j * rnI_I / 2

        pnE = np.abs(pnE_R) - 1j * np.abs(pnE_I) if stableEffectivePoles else np.abs(pnE_R) + 1j * pnE_I
        rnE = rnE_R + 1j * rnE_I

        if poleOrigin:
            pnI = np.hstack([0., pnI])
            rnI = np.hstack([-1j * np.abs(r0) / 2, rnI])

        poles = np.hstack([pnC, pnI, pnE])
        residues = np.hstack([rnC, rnI, rnE])

    elif fctType == "gdl":
        Hnr, G0, an, bn, s1n, Gn, wn, s2n, pnE_R, pnE_I, rnE_R, rnE_I = parameterList[2:-1]

        s1nn = s1n
        Gnn = 0.00005 + np.abs(Gn)
        wnn = 0.00005 + np.abs(wn)
        s2nn = s2n

        pnE = np.abs(pnE_R) - 1j * np.abs(pnE_I) if stableEffectivePoles else np.abs(pnE_R) + 1j * pnE_I
        rnE = rnE_R + 1j * rnE_I

        pC = wnn - 1j * Gnn / 2
        pI = np.hstack([0, -1j * np.abs(bn)])
        poles = np.hstack([pC, pI, pnE])

        resPoles = 1j * s1nn * Gnn * pC + s2nn * np.abs(pC) ** 2
        resPoles /= -2 * np.real(pC)
        resPolesIm = -1j * (an ** 2) / np.abs(bn)
        residues = np.hstack([resPoles, np.hstack([1j * G0 - resPolesIm.sum(), resPolesIm]) / 2, rnE])

    elif fctType == "mtse":
        Hnr, mrn, trn, wnR, wnI, pnE_R, pnE_I, rnE_R, rnE_I = parameterList[2:-1]

        rnC = np.abs(mrn) * np.exp(1j * trn)
        pnC = np.abs(wnR) - 1j * np.abs(wnI)

        pnE = np.abs(pnE_R) - 1j * np.abs(pnE_I) if stableEffectivePoles else np.abs(pnE_R) + 1j * pnE_I
        rnE = rnE_R + 1j * rnE_I

        poles = np.hstack([pnC, pnE])
        residues = np.hstack([rnC, rnE])

    else:
        print('Invalid fct type. Available options are "sem", "gdl" and "mtse" ')
        exit()

    return poles, residues


def GetPolesAndZeros(oldParameterList):
    poles, zeros = None, None

    fctType = oldParameterList[0]
    tensorized = oldParameterList[1]
    stableEffectivePoles = oldParameterList[-1]

    parameterList = oldParameterList if not tensorized else Numpify(oldParameterList)

    if fctType == "szf":
        (G0, pnC_R, pnC_I, pnI_I, pnE_R, pnE_I, znC_R, znC_I, znI_I, znE_R, znE_I,
         poleOrigin, reversibleZeros) = parameterList[2:-1]

        pnC = np.abs(pnC_R) - 1j * np.abs(pnC_I)
        znC = np.abs(znC_R) + 1j * znC_I if not reversibleZeros else np.abs(znC_R) - 1j * np.abs(znC_I)

        pnI = -1j * np.abs(pnI_I)
        znI = 1j * znI_I / 2 if (not reversibleZeros) else -1j * np.abs(znI_I) / 2

        if poleOrigin:
            pnI = np.hstack([0., pnI])

        pnE = np.abs(pnE_R) - 1j * np.abs(pnE_I) if stableEffectivePoles else np.abs(pnE_R) + 1j * pnE_I
        znE = np.abs(znE_R) - 1j * np.abs(znE_I) if reversibleZeros else np.abs(znE_R) + 1j * znE_I

        poles = np.hstack([pnC, pnI, pnE])
        zeros = np.hstack([znC, znI, znE])

    return poles, zeros


import numpy as np
import pandas as pd

c = 3e8
hb = 6.582119570e-1


def ReadDataFile(dataFilePath, columnNameIndex=None, FreqColName=None, DataColNameR=None, DataColNameI=None,
                 sep=",", decimal=".", useLambda=False, FreqConversion=1e-15, w1=-1, w2=500, nPts=75):
    a = pd.read_csv(dataFilePath, header=columnNameIndex, delimiter=sep, decimal=decimal)
    X = a.iloc[:, 0] if FreqColName is None else a[FreqColName]
    W = np.flip(2 * np.pi * c / X.to_numpy()).copy().astype(float) if useLambda else X.to_numpy().astype(float)
    W *= FreqConversion

    HwR = a.iloc[:, 1] if DataColNameR is None else a[DataColNameR]
    HwR = np.flip(HwR.to_numpy()).copy() if useLambda else HwR.to_numpy()
    HwI = a.iloc[:, 2] if DataColNameI is None else a[DataColNameI]
    HwI = np.flip(HwI.to_numpy()).copy() if useLambda else HwI.to_numpy()
    Hw = HwR + 1j * HwI

    Hw = Hw[np.argsort(W)]
    W = W[np.argsort(W)]

    Hw = Hw[(w1 < W) & (W < w2)]
    W = W[(w1 < W) & (W < w2)]

    if W.shape[0] > nPts:
        q = np.unique(np.round(np.linspace(0, W.shape[0] - 1, num=nPts)).astype(int))
        Hw = Hw[q]
        W = W[q]

    return W, Hw